from .is_iterable import isIterable
from .no_value import NoValue
from .extract_path_keys import extract_path_keys
from .get_path_array import get_path_array
from .is_flat import isFlat

__all__ = ['isIterable', 'NoValue']


from typing import Any


# def isIterable(obj: Any) -> bool:
#     try:
#         iter(obj)
#     except:
#         return False
#     else:
#         return True


def isIterable(obj: Any) -> bool:
    try:
        obj.__iter__()
    except:
        return False
    else:
        return True



from typing import Dict, TypeVar, Hashable, List, overload

from ..objects import chunkDict
from ..arrays import chunkList

V = TypeVar('V')
K = TypeVar('K', bound=Hashable)


@overload
def chunk(obj: List[V], size: int = 1) -> List[List[V]]:
    ...


@overload
def chunk(obj: Dict[K, V], size: int = 1) -> List[Dict[K, V]]:
    ...


def chunk(obj, size: int = 1):
    if isinstance(obj, dict):
        return chunkDict(obj, size)
    else:
        return chunkList(obj, size)



from .get import get
from .group_by import groupBy
from .head import head
from .at import at
from .chunk import chunk
from .set_ import set_
from .merge import merge
from .pick import pick, pickDeep

__all__ = ['get', 'groupBy', 'head', 'at', 'chunk', 'set_', 'merge', 'pick', 'pickDeep']



from typing import Union, Iterable, List, Callable, Any, Hashable, Dict, TypeVar, overload
from .get import get

V = TypeVar('V')
K = TypeVar('K', bound=Hashable)
Y = TypeVar('Y', bound=Hashable)


@overload
def groupBy(obj: Dict[K, V], iteratee: Union[Callable[[V], Y], Any]) -> Dict[Y, List[V]]:
    ...


@overload
def groupBy(obj: Iterable[V], iteratee: Union[Callable[[V], Y], Any] = None) -> Dict[Y, List[V]]:
    ...


def groupBy(obj: Union[Dict[Any, V], Iterable[V]],
            iteratee: Union[Callable[[V], Y], Any] = None) -> Dict[Y, List[V]]:
    res: Dict[Y, List[V]] = {}
    try:
        values = obj.values()
    except:
        values = obj

    if iteratee is None:
        for v in values:
            elems = res.setdefault(v, [])
            elems.append(v)
    else:
        if not callable(iteratee):
            it = iteratee

            def iteratee(x: V) -> Y:
                return get(x, it)
        for v in values:
            elems = res.setdefault(iteratee(v), [])
            elems.append(v)
    return res



from typing import Iterable, TypeVar, Iterable

from .is_iterable import isIterable

T = TypeVar('T')


def isFlat(array: Iterable[T]) -> bool:
    for x in array:
        if isIterable(x) and (not isinstance(x, str) or (x.find('.') != -1 or x.find('[') != -1)):
            return False
    return True



from typing import Dict, TypeVar, Hashable

V = TypeVar('V')
K = TypeVar('K', bound=Hashable)


def renameKeys(obj: Dict[K, V], keyMap: Dict[K, K]) -> Dict[K, V]:
    return {keyMap.get(k, k): v for k, v in obj.items()}



from typing import Iterable, List, TypeVar
from itertools import chain

from .uniq import uniq

T = TypeVar('T')


def union(array: Iterable[T], *others: Iterable[T]) -> List[T]:
    return uniq(chain(array, *others))



from typing import Union, List, Hashable, TypeVar, Dict, Callable, Optional

from ..collections import head
from ..tools import NoValue

V = TypeVar('V')
K = TypeVar('K', bound=Hashable)


def mapKeys(obj: Dict[K, V], iteratee: Callable[[V, Optional[K], Optional[Dict[K, V]]], K]) -> Dict[K, V]:
    n = _check_args_number(iteratee, obj)
    return {iteratee(*(v, k, obj)[:n]): v for k, v in obj.items()}


def _check_args_number(it, obj):
    k = head(obj, NoValue)
    if k is NoValue:
        return 3
    v = obj[k]
    try:
        it(v)
        return 1
    except:
        try:
            it(v, k)
            return 2
        except:
            return 3



from typing import Dict, TypeVar, Hashable, List

V = TypeVar('V')
K = TypeVar('K', bound=Hashable)


def chunkDict(obj: Dict[K, V], size: int = 1) -> List[Dict[K, V]]:
    res = []
    if size < 1:
        raise ValueError("size must be greater 0")
    ks = list(obj)
    while len(ks):
        res.append({k: obj[k] for k in ks[:size]})
        ks = ks[size:]
    return res



from typing import Union, List, Hashable, TypeVar

from ..tools import isIterable
from .extract_path_keys import extract_path_keys

Y = TypeVar('Y', Hashable, str, int)


def get_path_array(path: Union[Y, List[Y]]) -> List[Y]:
    if isinstance(path, str):
        return extract_path_keys(path)
    elif not isIterable(path):
        return [path]
    return path



from typing import Callable, Hashable, Iterable, TypeVar, List, Any

T = TypeVar('T')
X = TypeVar('X', bound=Hashable)


def uniq(array: Iterable[T]) -> List[T]:
    res: List[T] = []
    for v in array:
        if v not in res:
            res.append(v)
    return res


def uniqBy(array: Iterable[T], iteratee: Callable[[T], Any]) -> List[T]:
    res: List[T] = []
    keys: List[Any] = []
    for v in array:
        a = iteratee(v)
        if a not in keys:
            keys.append(a)
            res.append(v)
    return res


def uniqHash(array: Iterable[X]) -> List[X]:
    return list(set(array))


def uniqHashBy(array: Iterable[X], iteratee: Callable[[X], int]) -> List[X]:
    res = {iteratee(x): x for x in array}
    return list(res.values())



from typing import Iterable, TypeVar, Iterable, Hashable, Dict

from ..tools import NoValue


T = TypeVar('T')
V = TypeVar('V')
K = TypeVar('K', bound=Hashable)


def pickPlain(obj: Dict[K, V], keys: Iterable[K]) -> Dict[K, V]:
    res = {}
    for k in keys:
        if k in obj:
            res[k] = obj[k]
    return res



from .uniq import uniq, uniqBy, uniqHash, uniqHashBy
from .union import union
from .flatten import flatten, flattenDeep, flattenDepth
from .concat import concat
from .intersection import intersection
from .sum_by import sumBy
from .chunk_list import chunkList
from .group_by import groupListBy
from .find_index import findIndex
from .duplicates import duplicates, duplicatesHash

__all__ = ['uniq', 'union', 'flatten', 'concat', 'intersection', 'sumBy', 'chunkList', 'groupListBy',
           'flattenDeep', 'flattenDepth', 'findIndex', 'uniqBy', 'uniqHash', 'uniqHashBy', 'duplicates',
           'duplicatesHash']



from typing import List, Any, Hashable, TypeVar, Union
import copy

from .get import get
from ..tools import NoValue, get_path_array

T = TypeVar('T')


def set_(obj: T, path: Union[List[Hashable], Hashable], value: Any, deep_copy=False) -> T:
    path = get_path_array(path)
    if deep_copy:
        obj = copy.deepcopy(obj)

    current_dest = obj
    nxt = type(obj)
    for k, v in zip(path, path[1:]):
        if k == '':
            nxt = dict
            continue
        if v == '':
            nxt = list
        current_dest = _mergeObjStruct(current_dest, nxt, k)

    last = path[-1]
    _setBasic(current_dest, last, value)
    return obj


def _mergeObjStruct(dest: Any, typ: Any, key: Any) -> Any:
    ext = get(dest, key, NoValue)
    if ext is NoValue:
        value = typ()
    else:
        return ext

    _setBasic(dest, key, value)
    return value


def _setBasic(dest: Any, key: Any, value: Any) -> Any:
    try:
        dest[key] = value
    except:
        if hasattr(dest, 'extend'):
            dest.extend([None] * (key - len(dest)) + [value])
        elif isinstance(key, str):
            setattr(dest, key, value)
    return dest



from typing import Iterable, TypeVar, Iterable, Hashable, Dict
import copy


V = TypeVar('V')
K = TypeVar('K', bound=Hashable)


def omitPlain(obj: Dict[K, V], keys: Iterable[K]) -> Dict[K, V]:
    res = copy.copy(obj)
    for k in keys:
        if k in res:
            del res[k]
    return res



from typing import Any, Iterable, TypeVar, Union, overload, KeysView, Optional, Dict


T = TypeVar('T')
Y = TypeVar('Y')


@overload
def head(iterable: KeysView[T], default: Y) -> Union[T, Y]:
    ...


@overload
def head(iterable: Iterable[T], default: Y) -> Union[T, Y]:
    ...


@overload
def head(iterable: KeysView[T], default=None) -> Optional[T]:
    ...


@overload
def head(iterable: Iterable[T], default=None) -> Optional[T]:
    ...


def head(iterable: Union[Iterable[T], Dict[T, Any]], default: Y = None) -> Union[Optional[T], Y]:
    try:
        return next(iter(iterable), default)
    except:
        return default



import re

pat = re.compile(r'(?:[^\w\d]|_)+')


def camelCase(text: str) -> str:
    spl = pat.split(text)
    ns = ''.join(f'{x[:1].capitalize()}{x[1:]}' if x.upper() != x else x.capitalize() for x in spl)
    return f'{ns[:1].lower()}{ns[1:]}'



from typing import Iterable, TypeVar, List, Union

from ..tools import isIterable

V = TypeVar('V')


def flatten(array: Iterable[Union[V, Iterable[V]]]) -> List[V]:
    return flattenDepth(array)


def flattenDeep(array: Iterable[Union[V, Iterable[V]]]) -> List[V]:
    return flattenDepth(array, 0)


def flattenDepth(array: Iterable[Union[V, Iterable[V]]], depth=1) -> List[V]:
    res = []
    folded = False
    for x in array:
        if isIterable(x) and not isinstance(x, str):
            res.extend(x)
            folded = True
        else:
            res.append(x)
    if depth != 1 and folded:
        return flattenDepth(res, depth - 1)
    else:
        return res



import re
from typing import List, Union

pat = re.compile(r'(\[\d+\])')


def extract_path_keys(path: str) -> List[Union[str, int]]:
    res: List[Union[str, int]] = []
    for v in path.split('.'):
        if v == '':
            continue
        path_splitted = pat.split(v)
        if len(path_splitted) == 1:
            res.extend(path_splitted)
            continue
        r1: List[Union[str, int]] = [''] * len(path_splitted)
        r1[::2] = path_splitted[::2]
        r1[2::2] = [int(p[1:-1]) for p in path_splitted[1::2]]
        res.extend(r1)
    return res



from typing import Any, Dict, TypeVar, List, Hashable, Union, overload

from ..tools import isFlat, NoValue
from ..objects import pickPlain
from ..arrays import flatten
from ..collections import set_, get

T = TypeVar('T')
K = TypeVar('K', bound=Hashable)


@overload
def pick(obj: List[T], *properties) -> Dict[K, T]:
    ...


@overload
def pick(obj: Dict[K, T], *properties) -> Dict[K, T]:
    ...


def pick(obj: Union[List[T], Dict[K, T]], *properties) -> Dict[K, T]:
    propsFlaten = flatten(properties)
    if isinstance(obj, dict) and isFlat(propsFlaten):
        return pickPlain(obj, propsFlaten)
    else:
        return pickDeep(obj, propsFlaten)


def pickDeep(obj: Union[List[T], Dict[K, T]], paths: List[K], deepCopy=False) -> Dict[K, T]:
    res: Dict[K, T] = {}
    for path in paths:
        value = get(obj, path, NoValue, deepCopy)
        if value is not NoValue:
            set_(res, path, value)
    return res



from typing import Dict, Iterable, List, TypeVar, Callable, Hashable, Any

from ..tools.no_value import NoValue

V = TypeVar('V')
VH = TypeVar('VH', bound=Hashable)
K = TypeVar('K', bound=Hashable)


def duplicates(array: Iterable[V], iteratee: Callable[[V], Any] = None) -> List[V]:
    res: List[V] = []
    if iteratee is None:
        tmp = list(array)
    elif callable(iteratee):
        tmp = [iteratee(x) for x in array]
    else:
        raise ValueError('wrong iteratee')

    array = list(array)
    for i, v in enumerate(tmp):
        if v is NoValue:
            continue
        dups = []
        for ix, x in enumerate(tmp[i+1:]):
            if v == x:
                dups.append(array[i+1+ix])
                tmp[i+1+ix] = NoValue
        if dups:
            res.append(dups[-1])
    return res


def duplicatesHash(array: Iterable[VH], iteratee: Callable[[VH], K] = None) -> List[VH]:
    res: Dict[Hashable, List[VH]] = {}
    tmp: Dict[Hashable, int] = {}
    if iteratee is None:
        for v in array:
            if v not in tmp:
                tmp[v] = 1
                res[v] = []
            else:
                res[v].append(v)
    elif callable(iteratee):
        for v in array:
            k = iteratee(v)
            if k not in tmp:
                tmp[k] = 1
                res[k] = []
            else:
                res[k].append(v)
    return [x[-1] for x in res.values() if x]



from typing import Any, Iterable, TypeVar, Callable, Union

V = TypeVar('V')
D = TypeVar('D', None, Any)


def findIndex(array: Iterable[V], iteratee: Union[Callable[[V], bool], V], default: D = None) -> Union[int, D]:
    if callable(iteratee):
        for k, v in enumerate(array):
            if iteratee(v):
                return k
    else:
        for k, v in enumerate(array):
            if v == iteratee:
                return k
    return default



from typing import Any, Iterable, List, TypeVar, Iterable, Union, overload

from .get import get
from ..tools.no_value import NoValue

T = TypeVar('T')
K = TypeVar('K', str, int)


@overload
def at(obj: Iterable[Any], paths: Iterable[K]) -> List[Any]:
    ...


@overload
def at(obj: Iterable[Any], paths: Iterable[Iterable[K]]) -> List[Any]:
    ...


def at(obj: Any, paths: Iterable[Any]) -> List[Any]:
    return [get(obj, p) for p in paths]



from typing import TypeVar
import copy

from .get import get
from .set_ import _setBasic

T = TypeVar('T')


def merge(dest: T, src: T, maxDepth=-1, deepCopySrc=False) -> T:
    if maxDepth == 0:
        return src

    if deepCopySrc:
        src = copy.deepcopy(src)

    if isinstance(src, dict) and isinstance(dest, dict):
        iterate = src.items()
    elif isinstance(src, list) and isinstance(dest, list):
        iterate = enumerate(src)
    else:
        return src

    for key, src_value in iterate:
        dest_value = get(dest, [key])
        res_value = merge(dest_value, src_value, maxDepth=maxDepth-1)
        _setBasic(dest, key, res_value)
    return dest



from .pick import pickPlain
from .omit import omitPlain
from .chunk_dict import chunkDict
from .find_key import findKey
from .rename_keys import renameKeys
from .map_keys import mapKeys
from .invert import invert

__all__ = ['pickPlain', 'omitPlain', 'chunkDict', 'findKey', 'renameKeys', 'mapKeys', 'invert']



from typing import Iterable, List, TypeVar

from itertools import chain

T = TypeVar('T')


def intersection(array: Iterable[T], *others: Iterable[T]) -> List[T]:
    res: List[T] = []
    merged = [array, *others]
    if len(merged) == 1:
        return list(array)
    mn = min([(len(x), i) for i, x in enumerate(merged)], key=lambda x: x[0])
    a = merged.pop(mn[1])
    iters = chain(*merged)
    for v in iters:
        if v in a and v not in res:
            res.append(v)
    return res



from setuptools import setup, find_packages

setup(
    name='pydashlite',
    version='0.1.2',
    description='Simple tools similar to pydash, but more specific and faster.',
    url='https://github.com/glowlex/pydashlite',
    author='glowlex',
    author_email='glowlex@yandex.ru',
    license='MIT',
    packages=find_packages(exclude=['tests']),
    zip_safe=False,
    test_suite="tests",
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
    )



from typing import Callable, Iterable, TypeVar

T = TypeVar('T')
Num = TypeVar('Num', int, float)


def sumBy(array: Iterable[T], iteratee: Callable[[T], Num] = None, start: Num = 0) -> Num:
    if iteratee is None:
        return sum([y for y in array], start)
    return sum([iteratee(y) for y in array], start)



from typing import List, TypeVar

V = TypeVar('V')


def chunkList(array: List[V], size: int = 1) -> List[List[V]]:
    res = []
    if size < 1:
        raise ValueError("size must be greater 0")
    while len(array):
        res.append(array[:size])
        array = array[size:]
    return res



from typing import Any, Dict, TypeVar, Hashable, Callable, Union

V = TypeVar('V')
K = TypeVar('K', bound=Hashable)
D = TypeVar('D', None, Any)


def findKey(obj: Dict[K, V], iteratee: Union[Callable[[V], bool], V], default: D = None) -> Union[K, D]:
    if callable(iteratee):
        for k, v in obj.items():
            if iteratee(v):
                return k
    else:
        for k, v in obj.items():
            if v == iteratee:
                return k
    return default



from typing import Any, Hashable, List, Union, TypeVar
import copy

from ..tools import NoValue, get_path_array

Y = TypeVar('Y', Hashable, str, int)


def get(obj: Any, path: Union[Y, List[Y]], default=None, doDeepCopy=False) -> Any:
    res = obj
    for k in get_path_array(path):
        if k == '':
            continue
        try:
            res = res[k]
        except LookupError:
            return default
        except TypeError:
            try:
                res = getattr(res, k)
                if callable(res):
                    return default
            except:
                return default
    if doDeepCopy:
        return copy.deepcopy(res)
    else:
        return res
